import time
import anyio
import moat.lib.gpio as gpio

"""
This example is taken out of my furnace controller.
It has been tested with Raspberry Pi Zero W and I assume it will work with any board supported by moat.gpio.
Use at your own risk.

If you aren't sure about how to hook up a button and led to your board, there are a lot of examples online.

Thank you @smurfix, who wrote moat.lib.gpio and @njsmith and other in glitter:python-trio/general room
who helped me out.
"""


class Led:
    # This class turns on and off the power to a pin.
    # Two events are setup for turning off an on the pin. Both events need to be
    # called at the same time or trio might await at the wrong spot.
    def __init__(self, line):
        self.x = line
        self._on = anyio.create_event()
        self._off = anyio.create_event()

    async def liteon(self):
        with gpio.open_chip() as chip:
            with chip.line(self.x).open(direction=gpio.DIRECTION_OUTPUT) as line:
                self._on.clear()
                await self._off.set()
                while True:
                    if self._on.is_set():
                        line.value = 1
                        # print('lite on')
                        await self._off.wait()
                        self._on = anyio.create_event()
                    elif self._off.is_set():
                        line.value = 0
                        # print('lite off')d
                        await self._on.wait()
                        self._off = anyio.create_event()
                    else:
                        # should never be reached.
                        # if the code does reach here,
                        # turn off the power to whatever is being powered
                        print("error: both are off.")
                        await self._off.set()


class Button:
    # Add the events tthe button is attached to and the on off event are passed into the class.
    # The class listens for the voltage to rise then reverses whatever the current settings are.
    def __init__(self, line, event_on, event_off):
        self.y = line
        self._on = event_on
        self._off = event_off

    async def push(self):
        with gpio.Chip(0) as c:
            in_ = c.line(self.y)
            with in_.monitor(gpio.REQUEST_EVENT_RISING_EDGE):
                last = 0
                async for e in in_:
                    # This section is for debouncing the button.
                    # As a button is pushed and released the voltage can rapidly go up and down many times
                    # when the user only meant one push. To limit this, a delay is add to ignore changes.
                    # This can be adjusted depending on the button and the respose.
                    secs, ns_secs = e.timestamp
                    now = float(str(secs) + "." + str(ns_secs))
                    if now >= last + 0.25:
                        print("button", e.value, secs, ns_secs, now)
                        if self._on.is_set():
                            await self._off.set()
                        else:
                            await self._on.set()
                    last = now


# Asyncgpio uses the BCM pin numbering. So, the led is on the pin 21
# and the button that controls the yellow is hooked to pin 23.
yellow = Led(21)
yellowbutton = Button(23, yellow._on, yellow._off)


async def main(y):
    async with anyio.create_task_group() as nursery:
        await nursery.spawn(yellowbutton.push)
        await nursery.spawn(yellow.liteon)


if __name__ == "__main__":
    anyio.run(main, 1, backend="trio")



import anyio
import moat.lib.gpio as gpio

"""
This script toggles a pin and watches another. The two are presumed to be connected (hardware wire).
"""


async def pling(line):
    while True:
        await anyio.sleep(1)
        line.value = 1
        await anyio.sleep(1)
        line.value = 0


async def main():
    async with anyio.create_task_group() as n:
        with gpio.Chip(0) as c:
            with c.line(19).open(direction=gpio.DIRECTION_OUTPUT) as out_:
                in_ = c.line(20)
                await n.spawn(pling, out_)
                with in_.monitor(gpio.REQUEST_EVENT_BOTH_EDGES):
                    async for e in in_:
                        print(
                            e,
                            "on" if e.value else "off",
                            "at",
                            e.time.strftime("%H:%M:%S"),
                        )


if __name__ == "__main__":
    anyio.run(main, backend="trio")



import moat.lib.gpio as gpio
import time

"""Flash an output manually.

On the Pi3, control the LEDs thus:
# cd /sys/class/leds/led0 ## or /led1
# echo gpio >trigger
# echo 16 >gpio

Enjoy.

NB: The red LED can go much faster. Have fun.
The green LED cannot (limited by hardware,
so that you can still see very fast flashes)

"""
if __name__ == "__main__":
    with gpio.Chip(0) as c:
        with c.line(16).open(gpio.DIRECTION_OUTPUT) as l:
            try:
                while True:
                    l.value = 1
                    time.sleep(0.1)
                    l.value = 0
                    time.sleep(0.1)
            finally:
                l.value = 0



"""
This module contains helpers for testing async gpio, via the Linux kernel's
``gpio_mockup`` module (writing) and ``/sys/kernel/debug/cpio`` (monitoring).
"""

from __future__ import annotations

import errno
import logging
import os
import re
from collections import namedtuple
from contextlib import asynccontextmanager

import anyio
from moat.util import Queue

logger = logging.getLogger(__name__)

_r_chip = re.compile(
    "^(?P<chip>[a-z0-9]+): GPIOs (?P<base>[0-9]+)-(?:.*, (?P<name>[-_a-zA-Z0-9]+): *$)?",
)
_r_pin = re.compile("^gpio-(?P<pin>[0-9]+) \\(.*\\) (?P<dir>in|out) +(?P<val>hi|lo)")

Pin = namedtuple("Pin", ["out", "level"])


class _GpioPin:
    """
    Code representing one GPIO pin.
    """

    fd = None

    def __init__(self, watcher: GpioWatcher, chip: str, pin: int):
        self.watcher = watcher
        self.chip = chip
        self.pin = pin
        self.mon = set()
        self.state = (None, None)
        try:
            self.fd = os.open(
                os.path.join(watcher.debugfs_path, "gpio-mockup-event", chip, str(pin)),
                os.O_WRONLY,
            )
        except OSError as exc:
            if exc.errno != errno.ENOENT:
                raise

    def __del__(self):
        if self.fd is not None:
            os.close(self.fd)
            del self.fd

    @asynccontextmanager
    async def watch(self):
        """
        An async context manager that returns an iterator for changes of
        this pin.

        Values are (out,level) tuples of bool, with "out" and "high"
        represented as True.
        """
        q = Queue(10)
        self.mon.add(q)
        try:
            yield q
        finally:
            self.mon.remove(q)

    async def see(self, write: bool, level: bool):
        s = (write, level)
        if self.state == s:
            return
        self.state = s
        logger.debug("SEE %s %d %s", self.chip, self.pin, self.state)
        for cb in list(self.mon):
            await cb.put(s)

    def set(self, value: bool):
        logger.debug("SET %s %d %s", self.chip, self.pin, value)
        if self.fd is None:
            raise RuntimeError(
                f"Pin {self.chip}/{self.pin} is not controlled via the 'gpio_mockup' module",
            )
        os.write(self.fd, b"1" if value else b"0")
        # os.lseek(self.fd, 0, os.SEEK_SET)


class GpioWatcher:
    """
    Code which triggers callbacks whenever a GPIO pin changes.

    This class polls `/sys/kernel/debug/gpio` (can be overridden).
    """

    tg = None  # for .run

    def __init__(
        self,
        interval: float = 0.2,
        debugfs_path: str = "/sys/kernel/debug",
        sysfs_path: str = "/sys",
    ):
        self.interval = interval
        self.gpio = open(  # pylint: disable=unspecified-encoding
            os.path.join(debugfs_path, "gpio"),
        )
        self.targets = dict()  # chip > line > _GpioPin
        #       self.names = {}
        self.sysfs_path = sysfs_path
        self.debugfs_path = debugfs_path
        # gpio_dir = os.path.join(sysfs_path, "class", "gpio")

    #       for d in os.listdir(gpio_dir):
    #           try:
    #               with open(os.path.join(gpio_dir,d,"label"),"r") as f:
    #                   n = f.read().strip()
    #           except EnvironmentError as e:
    #               if e.errno == errno.ENOTDIR:
    #                   continue
    #               raise
    #           else:
    #               self.names[d] = n

    def monitor(self, chip: str, pin: int):
        """
        Shortcut for 'self.pin(chip, pin).watch()'.
        """
        return self.pin(chip, pin).watch()

    def pin(self, chip: str, pin: int, create: bool = True):
        """
        Returns a pins corresponding GpioPin
        """
        #       chip = self.names[chip]
        try:
            c = self.targets[chip]
        except KeyError:
            if not create:
                raise
            self.targets[chip] = c = dict()
        try:
            p = c[pin]
        except KeyError:
            if not create:
                raise
            c[pin] = p = _GpioPin(self, chip, pin)
        return p

    async def _watch(self):
        # The actual monitor.
        while True:
            await self.check_pins()
            await anyio.sleep(self.interval)

    async def check_pins(self):
        """
        Read the GPIO debug file and update pin states
        """
        chip = None
        base = None

        for line in self.gpio:
            line = line.strip()
            if not line:
                chip = None
                continue
            if chip is None:
                r = _r_chip.match(line)
                if not r:
                    raise ValueError(line)
                chip = r.group("name")
                if not chip:
                    chip = r.group("chip")
                base = int(r.group("base"))
            else:
                r = _r_pin.match(line)
                if not r:
                    raise ValueError(line)
                pin = int(r.group("pin")) - base
                out = r.group("dir") == "out"
                val = r.group("val") == "hi"

                try:
                    pin = self.pin(chip, pin, create=False)
                except KeyError:
                    pass
                else:
                    await pin.see(out, val)
        self.gpio.seek(0)

    @asynccontextmanager
    async def run(self):
        """
        This async context manager controls the monitoring loop.
        """
        async with anyio.create_task_group() as tg:
            self.tg = tg
            await tg.spawn(self._watch)
            try:
                yield self
            finally:
                self.tg = None
                await tg.cancel_scope.cancel()



# Copyright (c) 2018 Matthias Urlichs <matthias@urlichs.de>
# Based on work Copyright (c) 2018 Steven P. Goldsmith
"""
libgpiod CFFI interface
-------------

This is a stripped-down version which doesn't do "bulk" access (no point IMHO)
and doesn't implement an event loop (that's Trio's job).
"""

from __future__ import annotations

from cffi import FFI

__all__ = [
    "DIRECTION_INPUT",
    "DIRECTION_OUTPUT",
    "ACTIVE_STATE_HIGH",
    "ACTIVE_STATE_LOW",
    "REQUEST_DIRECTION_AS_IS",
    "REQUEST_DIRECTION_INPUT",
    "REQUEST_DIRECTION_OUTPUT",
    "REQUEST_EVENT_FALLING_EDGE",
    "REQUEST_EVENT_RISING_EDGE",
    "REQUEST_EVENT_BOTH_EDGES",
    "REQUEST_FLAG_OPEN_DRAIN",
    "REQUEST_FLAG_OPEN_SOURCE",
    "REQUEST_FLAG_ACTIVE_LOW",
    "EVENT_RISING_EDGE",
    "EVENT_FALLING_EDGE",
    "ffi",
    "lib",
]

ffi = FFI()
ffi.cdef(
    """
enum {
    GPIOD_CTXLESS_EVENT_CB_TIMEOUT = 1,
    GPIOD_CTXLESS_EVENT_CB_RISING_EDGE,
    GPIOD_CTXLESS_EVENT_CB_FALLING_EDGE,
};

enum {
    GPIOD_CTXLESS_EVENT_CB_RET_ERR = -1,
    GPIOD_CTXLESS_EVENT_CB_RET_OK = 0,
    GPIOD_CTXLESS_EVENT_CB_RET_STOP = 1,
};

enum {
    GPIOD_CTXLESS_EVENT_POLL_RET_STOP = -2,
    GPIOD_CTXLESS_EVENT_POLL_RET_ERR = -1,
    GPIOD_CTXLESS_EVENT_POLL_RET_TIMEOUT = 0,
};

enum {
    GPIOD_LINE_DIRECTION_INPUT = 1,
    GPIOD_LINE_DIRECTION_OUTPUT,
};

enum {
    GPIOD_LINE_ACTIVE_STATE_HIGH = 1,
    GPIOD_LINE_ACTIVE_STATE_LOW,
};

enum {
    GPIOD_LINE_REQUEST_DIRECTION_AS_IS = 1,
    GPIOD_LINE_REQUEST_DIRECTION_INPUT,
    GPIOD_LINE_REQUEST_DIRECTION_OUTPUT,
    GPIOD_LINE_REQUEST_EVENT_FALLING_EDGE,
    GPIOD_LINE_REQUEST_EVENT_RISING_EDGE,
    GPIOD_LINE_REQUEST_EVENT_BOTH_EDGES,
};

enum {
    GPIOD_LINE_REQUEST_FLAG_OPEN_DRAIN = 1,
    GPIOD_LINE_REQUEST_FLAG_OPEN_SOURCE = 2,
    GPIOD_LINE_REQUEST_FLAG_ACTIVE_LOW = 4,
    GPIOD_LINE_REQUEST_FLAG_BIAS_DISABLE = 8,
    GPIOD_LINE_REQUEST_FLAG_BIAS_PULL_DOWN = 16,
    GPIOD_LINE_REQUEST_FLAG_BIAS_PULL_UP = 32,
};

enum {
    GPIOD_LINE_EVENT_RISING_EDGE = 1,
    GPIOD_LINE_EVENT_FALLING_EDGE,
};

struct timespec {
    long tv_sec;
    long tv_nsec;
};

struct gpiod_line {
    unsigned int offset;
    int direction;
    int active_state;
    bool used;
    bool open_source;
    bool open_drain;
    int state;
    bool up_to_date;
    struct gpiod_chip *chip;
    int fd;
    char name[32];
    char consumer[32];
};

struct gpiod_chip {
    struct gpiod_line **lines;
    unsigned int num_lines;
    int fd;
    char name[32];
    char label[32];
};

struct gpiod_ctxless_event_poll_fd {
    int fd;
    /**< File descriptor number. */
    bool event;
    /**< Indicates whether an event occurred on this file descriptor. */
};

struct gpiod_line_request_config {
    const char *consumer;
    int request_type;
    int flags;
};

struct gpiod_line_event {
    struct timespec ts;
    int event_type;
};

struct gpiod_chip;

struct gpiod_line;

struct gpiod_chip_iter;

struct gpiod_line_iter;

struct gpiod_line_bulk;

typedef void (*gpiod_ctxless_set_value_cb)(void *);

typedef int (*gpiod_ctxless_event_handle_cb)(int, unsigned int,
                            const struct timespec *, void *);

typedef int (*gpiod_ctxless_event_poll_cb)(unsigned int,
                struct gpiod_ctxless_event_poll_fd *,
                const struct timespec *, void *);

int gpiod_ctxless_set_value(const char *device, unsigned int offset, int value,
                bool active_low, const char *consumer,
                gpiod_ctxless_set_value_cb cb,
                void *data);

int gpiod_ctxless_set_value_multiple(const char *device,
                        const unsigned int *offsets,
                        const int *values, unsigned int num_lines,
                        bool active_low, const char *consumer,
                        gpiod_ctxless_set_value_cb cb,
                        void *data);

int gpiod_ctxless_find_line(const char *name, char *chipname,
                size_t chipname_size,
                unsigned int *offset);

int gpiod_chip_find_lines(struct gpiod_chip *chip, const char **names,
                         struct gpiod_line_bulk *bulk);

struct gpiod_chip *gpiod_chip_open(const char *path);

struct gpiod_chip *gpiod_chip_open_by_name(const char *name);

struct gpiod_chip *gpiod_chip_open_by_number(unsigned int num);

struct gpiod_chip *gpiod_chip_open_by_label(const char *label);

struct gpiod_chip *gpiod_chip_open_lookup(const char *descr);

void gpiod_chip_close(struct gpiod_chip *chip);

const char *gpiod_chip_name(struct gpiod_chip *chip);

const char *gpiod_chip_label(struct gpiod_chip *chip);

unsigned int gpiod_chip_num_lines(struct gpiod_chip *chip);

struct gpiod_line *
gpiod_chip_get_line(struct gpiod_chip *chip, unsigned int offset);

int gpiod_chip_get_lines(struct gpiod_chip *chip,
                         unsigned int *offsets, unsigned int num_offsets,
                         struct gpiod_line_bulk *bulk);

int gpiod_chip_get_all_lines(struct gpiod_chip *chip,
                             struct gpiod_line_bulk *bulk);

struct gpiod_line *
gpiod_chip_find_line(struct gpiod_chip *chip, const char *name);

unsigned int gpiod_line_offset(struct gpiod_line *line);

const char *gpiod_line_name(struct gpiod_line *line);

const char *gpiod_line_consumer(struct gpiod_line *line);

int gpiod_line_direction(struct gpiod_line *line);

int gpiod_line_active_state(struct gpiod_line *line);

bool gpiod_line_is_used(struct gpiod_line *line);

bool gpiod_line_is_open_drain(struct gpiod_line *line);

bool gpiod_line_is_open_source(struct gpiod_line *line);

int gpiod_line_update(struct gpiod_line *line);

bool gpiod_line_needs_update(struct gpiod_line *line);

int gpiod_line_request(struct gpiod_line *line,
                const struct gpiod_line_request_config *config,
                int default_val);

int gpiod_line_request_input(struct gpiod_line *line,
                    const char *consumer);

int gpiod_line_request_output(struct gpiod_line *line,
                    const char *consumer, int default_val);

int gpiod_line_request_rising_edge_events(struct gpiod_line *line,
                        const char *consumer);

int gpiod_line_request_falling_edge_events(struct gpiod_line *line,
                        const char *consumer);

int gpiod_line_request_both_edges_events(struct gpiod_line *line,
                        const char *consumer);

int gpiod_line_request_input_flags(struct gpiod_line *line,
                    const char *consumer, int flags);

int gpiod_line_request_output_flags(struct gpiod_line *line,
                    const char *consumer, int flags,
                    int default_val);

int gpiod_line_request_rising_edge_events_flags(struct gpiod_line *line,
                        const char *consumer,
                        int flags);

int gpiod_line_request_falling_edge_events_flags(struct gpiod_line *line,
                            const char *consumer,
                            int flags);

int gpiod_line_request_both_edges_events_flags(struct gpiod_line *line,
                            const char *consumer,
                            int flags);

void gpiod_line_release(struct gpiod_line *line);

bool gpiod_line_is_requested(struct gpiod_line *line);

bool gpiod_line_is_free(struct gpiod_line *line);

int gpiod_line_get_value(struct gpiod_line *line);

int gpiod_line_set_value(struct gpiod_line *line, int value);

int gpiod_line_set_value_bulk(struct gpiod_line_bulk *bulk,
                              const int *values);

int gpiod_line_event_wait(struct gpiod_line *line,
                const struct timespec *timeout);

int gpiod_line_event_read(struct gpiod_line *line,
                struct gpiod_line_event *event);

int gpiod_line_event_get_fd(struct gpiod_line *line);

int gpiod_line_event_read_fd(int fd, struct gpiod_line_event *event);

struct gpiod_line *
gpiod_line_get(const char *device, unsigned int offset);

struct gpiod_line *gpiod_line_find(const char *name);

void gpiod_line_close_chip(struct gpiod_line *line);

struct gpiod_chip *gpiod_line_get_chip(struct gpiod_line *line);

struct gpiod_chip_iter *gpiod_chip_iter_new(void);

void gpiod_chip_iter_free(struct gpiod_chip_iter *iter);

void gpiod_chip_iter_free_noclose(struct gpiod_chip_iter *iter);

struct gpiod_chip *
gpiod_chip_iter_next(struct gpiod_chip_iter *iter);

struct gpiod_chip *
gpiod_chip_iter_next_noclose(struct gpiod_chip_iter *iter);

struct gpiod_line_iter *
gpiod_line_iter_new(struct gpiod_chip *chip);

void gpiod_line_iter_free(struct gpiod_line_iter *iter);

struct gpiod_line *
gpiod_line_iter_next(struct gpiod_line_iter *iter);

const char *gpiod_version_string(void);
""",
)

try:
    lib = ffi.dlopen("libgpiod.so.2")
except OSError:
    lib = ffi.dlopen("c")  # workaround if we're only building docs

DIRECTION_INPUT = lib.GPIOD_LINE_REQUEST_DIRECTION_INPUT
DIRECTION_OUTPUT = lib.GPIOD_LINE_REQUEST_DIRECTION_OUTPUT
ACTIVE_STATE_HIGH = lib.GPIOD_LINE_ACTIVE_STATE_HIGH
ACTIVE_STATE_LOW = lib.GPIOD_LINE_ACTIVE_STATE_LOW
REQUEST_DIRECTION_AS_IS = lib.GPIOD_LINE_REQUEST_DIRECTION_AS_IS
REQUEST_DIRECTION_INPUT = lib.GPIOD_LINE_REQUEST_DIRECTION_INPUT
REQUEST_DIRECTION_OUTPUT = lib.GPIOD_LINE_REQUEST_DIRECTION_OUTPUT
REQUEST_EVENT_FALLING_EDGE = lib.GPIOD_LINE_REQUEST_EVENT_FALLING_EDGE
REQUEST_EVENT_RISING_EDGE = lib.GPIOD_LINE_REQUEST_EVENT_RISING_EDGE
REQUEST_EVENT_BOTH_EDGES = lib.GPIOD_LINE_REQUEST_EVENT_BOTH_EDGES

REQUEST_FLAG_OPEN_DRAIN = lib.GPIOD_LINE_REQUEST_FLAG_OPEN_DRAIN
REQUEST_FLAG_OPEN_SOURCE = lib.GPIOD_LINE_REQUEST_FLAG_OPEN_SOURCE
REQUEST_FLAG_ACTIVE_LOW = lib.GPIOD_LINE_REQUEST_FLAG_ACTIVE_LOW
EVENT_RISING_EDGE = lib.GPIOD_LINE_EVENT_RISING_EDGE
EVENT_FALLING_EDGE = lib.GPIOD_LINE_EVENT_FALLING_EDGE



from __future__ import annotations
import datetime
import sys

import anyio

from . import libgpiod as gpio


class Chip:
    """Represents a GPIO chip.

    Arguments:
        label: Chip label. Run "gpiodetect" to list GPIO chip labels.

        num: Chip number. Deprecated. Defaults to zero.
            Only used if you don't use a label.

        consumer: A string for display by kernel utilities.
            Defaults to the program name.

    """

    _chip = None

    def __init__(self, num=None, label=None, consumer=sys.argv[0]):
        self._num = num
        self._label = label
        if (num is None) == (label is None):
            raise ValueError("Specify either label or num")
        self._consumer = consumer

    def __repr__(self):
        if self._label is None:
            return "%s(%d)" % (self.__class__.__name__, self._num)
        else:
            return f"{self.__class__.__name__}({self._label})"

    def __enter__(self):
        if self._label is None:
            self._chip = gpio.lib.gpiod_chip_open_by_number(self._num)
        else:
            self._chip = gpio.lib.gpiod_chip_open_by_label(self._label.encode("utf-8"))
        if self._chip == gpio.ffi.NULL:
            raise OSError("unable to open chip")
        return self

    def __exit__(self, *tb):
        gpio.lib.gpiod_chip_close(self._chip)
        self._chip = None

    def line(self, offset, consumer=None):
        """Get a descriptor for a single GPIO line.

        Arguments:
            offset: GPIO number within this chip. No default.
            consumer: override the chip's consumer, if required.
        """
        if consumer is None:
            consumer = self._consumer
        return Line(self, offset, consumer=consumer)


_FREE = 0
_PRE_IO = 1
_IN_IO = 2
_PRE_EV = 3
_IN_EV = 4
_IN_USE = {_IN_IO, _IN_EV}


class Line:
    """Represents a single GPIO line.

    Create this object by calling :meth:`Chip.line`.
    """

    _line = None
    _direction = None
    _default = None
    _flags = None
    _ev_flags = None
    _state = _FREE

    _type = None

    def __init__(self, chip, offset, consumer=sys.argv[0][:-3]):
        self._chip = chip
        self._offset = offset
        self._consumer = consumer.encode("utf-8")
        self.__consumer = gpio.ffi.new("char[]", self._consumer)

    def __repr__(self):
        return "<%s %s:%d %s=%d>" % (
            self.__class__.__name__,
            self._chip,
            self._offset,
            self._line,
            self._state,
        )

    def open(self, direction=gpio.DIRECTION_INPUT, default=False, flags=0):
        """
        Create a context manager for controlling this line's input or output.

        Arguments:
            direction: input or output. Default: gpio.DIRECTION_INPUT.
            flags: to request pull-up/down resistors or open-collector outputs.

        Example::
            with gpio.Chip(0) as chip:
                line = chip.line(16)
                with line.open(direction=gpio.DIRECTION_INPUT) as wire:
                    print(wire.value)
        """
        if self._state in _IN_USE:
            raise OSError("This line is already in use")
        self._direction = direction
        self._default = default
        self._flags = flags
        self._state = _PRE_IO
        return self

    def __enter__(self):
        """Context management for use with :meth:`open` and :meth:`monitor`."""
        if self._state in _IN_USE:
            raise OSError("This line is already in use")
        if self._state == _FREE:
            raise RuntimeError("You need to call .open() or .monitor()")
        self._line = gpio.lib.gpiod_chip_get_line(self._chip._chip, self._offset)
        if self._line == gpio.ffi.NULL:
            raise OSError("unable to get line")

        if self._state == _PRE_IO:
            self._enter_io()
        elif self._state == _PRE_EV:
            self._enter_ev()
        else:
            raise RuntimeError("wrong state", self)
        return self

    def _enter_io(self):
        if self._direction == gpio.DIRECTION_INPUT:
            r = gpio.lib.gpiod_line_request_input_flags(self._line, self._consumer, self._flags)
        elif self._direction == gpio.DIRECTION_OUTPUT:
            r = gpio.lib.gpiod_line_request_output_flags(
                self._line,
                self._consumer,
                self._flags,
                self._default,
            )
        else:
            self.__exit__()
            raise RuntimeError("Unknown direction")
        if r != 0:
            self.__exit__()
            raise OSError("unable to set direction")
        self._state = _IN_IO
        return self

    def _enter_ev(self):
        req = gpio.ffi.new("struct gpiod_line_request_config*")
        req.consumer = self.__consumer
        req.request_type = self._type
        req.flags = self._flags
        if gpio.lib.gpiod_line_request(self._line, req, 0) != 0:
            raise OSError("unable to request event monitoring")
        self._state = _IN_EV

    def __exit__(self, *tb):
        if self._line is not None:
            try:
                gpio.lib.gpiod_line_release(self._line)
            finally:
                self._line = None
        self._state = _FREE

    def _is_open(self):
        if self._state not in _IN_USE:
            raise RuntimeError("Line is not open", self)

    @property
    def value(self):
        self._is_open()
        return gpio.lib.gpiod_line_get_value(self._line)

    @value.setter
    def value(self, value):
        self._is_open()
        gpio.lib.gpiod_line_set_value(self._line, value)

    @property
    def direction(self):
        if self._line is None:
            return self._direction
        return gpio.lib.gpiod_line_direction(self._line)

    @property
    def active_state(self):
        self._is_open()
        return gpio.lib.gpiod_line_active_state(self._line)

    @property
    def is_open_drain(self):
        self._is_open()
        return gpio.lib.gpiod_line_is_open_drain(self._line)

    @property
    def is_open_source(self):
        self._is_open()
        return gpio.lib.gpiod_line_is_open_source(self._line)

    @property
    def is_used(self):
        self._is_open()
        return gpio.lib.gpiod_line_is_used(self._line)

    @property
    def offset(self):
        if self._line is None:
            return self._offset
        return gpio.lib.gpiod_line_offset(self._line)

    @property
    def name(self):
        self._is_open()
        n = gpio.lib.gpiod_line_name(self._line)
        if n == gpio.ffi.NULL:
            return None
        return n

    @property
    def consumer(self):
        if self._line is None:
            return self._consumer
        n = gpio.lib.gpiod_line_consumer(self._line)
        if n == gpio.ffi.NULL:
            return None
        return gpio.ffi.string(n).decode("utf-8")

    def monitor(self, type=gpio.REQUEST_EVENT_RISING_EDGE, flags=0):  # pylint: disable=redefined-builtin
        """
        Monitor events.

        Arguments:
            type: which edge to monitor
            flags: REQUEST_FLAG_* values (ORed)

        Usage::

            with gpio.Chip(0) as chip:
                line = chip.line(13)
                with line.monitor():
                    async for event in line:
                        print(event)
        """
        if self._state in _IN_USE:
            raise OSError("This line is already in use")
        self._state = _PRE_EV
        self._type = type
        self._flags = flags
        return self

    def _update(self):
        self._is_open()
        if gpio.lib.gpiod_line_update(self._line) == -1:
            raise OSError("unable to update state")

    def __iter__(self):
        raise RuntimeError("You need to use 'async for', not 'for'")

    async def __aenter__(self):
        raise RuntimeError("You need to use 'with', not 'async with'")

    async def __aexit__(self, *_):
        raise RuntimeError("You need to use 'with', not 'async with'")

    def __aiter__(self):
        if self._state != _IN_EV:
            raise RuntimeError("You need to call 'with LINE.monitor() / async for event in LINE'")
        return self

    async def __anext__(self):
        if self._state != _IN_EV:
            raise RuntimeError("wrong state")

        ev = gpio.ffi.new("struct gpiod_line_event*")
        fd = gpio.lib.gpiod_line_event_get_fd(self._line)
        if fd < 0:
            raise OSError("line is closed")
        await anyio.wait_socket_readable(fd)
        self._is_open()
        r = gpio.lib.gpiod_line_event_read_fd(fd, ev)
        if r != 0:
            raise OSError("unable to read update")
        return Event(ev)

    async def aclose(self):
        """close the iterator."""
        pass


class Event:
    """Store a Pythonic representation of an event"""

    def __init__(self, ev):
        if ev.event_type == gpio.EVENT_RISING_EDGE:
            self.value = 1
        elif ev.event_type == gpio.EVENT_FALLING_EDGE:
            self.value = 0
        else:
            raise RuntimeError("Unknown event type")
        self._ts_sec = ev.ts.tv_sec
        self._ts_nsec = ev.ts.tv_nsec

    @property
    def timestamp(self):
        """Return a (second,nanosecond) tuple for fast timestamping"""
        return (self._ts_sec, self._ts_nsec)

    @property
    def time(self):
        """Return the event's proper datetime"""
        return datetime.datetime.fromtimestamp(self._ts_sec + self._ts_nsec / 1000000000)

    def __repr__(self):
        return f"<{self.value} @{self.time}>"



import anyio
import moat.lib.gpio as gpio

"""
This script toggles a pin and watches another. The two are presumed to be connected (hardware wire).
"""


async def pling(line):
    while True:
        line.value = 1
        await anyio.sleep(1)
        line.value = 0
        await anyio.sleep(1)


async def main():
    async with anyio.create_task_group() as n:
        with gpio.Chip(0) as c:
            with c.line(19).open(direction=gpio.DIRECTION_OUTPUT) as out_, c.line(
                20
            ).open(direction=gpio.DIRECTION_INPUT) as in_:
                await n.spawn(pling, out_)
                while True:
                    print(in_.value)
                    await anyio.sleep(0.3)


if __name__ == "__main__":
    anyio.run(main, backend="trio")
